#  Copyright (c) 2010 Vitja Makarov
#
#  Author: Vitja Makarov <vitja.makarov@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['CommandOptionParser', 'make_option', 'register_command']

import sys
from optparse import OptionParser, make_option

class CommandOptionParser(OptionParser):
    def __init__(self, commands=None, *args, **kwargs):
        if commands:
            self.commands = commands
        else:
            self.commands = {}
        self.current_cmd = None
        OptionParser.__init__(self, *args, **kwargs)
        self.usage = '%prog <command> [options] [args]'

    def add_command(self, name, func):
        self.commands[name] = func

    def set_command(self, cmd_name):
        try:
            cmd = self.commands[cmd_name]
        except KeyError:
            self.unknown_command(cmd_name)
        if hasattr(cmd, 'description'):
            self.description = cmd.description
        if hasattr(cmd, 'option_list'):
            if cmd.option_list:
                self.add_options(cmd.option_list)
        if hasattr(cmd, 'usage') and cmd.usage is not None:
            self.usage = "%%prog %s %s" % (cmd_name, cmd.usage)
        else:
            self.usage = "%%prog %s [options] [args]" % cmd_name
        self.current_cmd = cmd

    def format_help(self, formater=None):
        head = OptionParser.format_help(self, formater)
        parts = [head]
        if not self.current_cmd:
            parts += ["Type '%s --help <command>' for help on a specific subcommand.\n" % self.get_prog_name()]
            parts += ["Available commands are:"]
            cmds = self.commands.keys()
            cmds.sort()
            parts.extend(map(lambda x: '    %s' % x, cmds))
            parts.append('')
        return '\n'.join(parts)

    def unknown_command(self, cmd):
        self.exit(1, "Unknown command '%s'\nType '%s --help' for usage\n" % (cmd, self.get_prog_name()))

    def parse_args(self, args=None):
        if args is None:
            args = sys.argv[1:]
        for i in xrange(len(args)):
            if args[i].startswith('-'):
                continue
            break
        else:
            self.print_help(sys.stderr)
            self.exit(1)
        cmd_name = args[i]
        args.pop(i)
        self.set_command(cmd_name)
        values, rest = OptionParser.parse_args(self, args)

        if (self.current_cmd.args_max and len(rest) > self.current_cmd.args_max) or \
            (self.current_cmd.args_min and len(rest) < self.current_cmd.args_min):
            self.print_help(sys.stderr)
            self.exit(1)
        return self.current_cmd, values, rest

def register_command(obj, name, option_list=None, description=None, usage=None, args_min=None, args_max=None):
    def wrapper(func):
        def decorator(*args, **kwargs):
            return func(*args, **kwargs)
        decorator.name = name
        decorator.option_list = option_list
        decorator.description = description
        decorator.usage = usage
        decorator.args_min = args_min
        decorator.args_max = args_max
        obj.add_command(name, decorator)
        return decorator
    return wrapper
