#!/usr/bin/env
#  Copyright (c) 2010 Vitja Makarov
#
#  Author: Vitja Makarov <vitja.makarov@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os.path

from extractiso import ExtractISO
from buildiso import BuildISO
from squashfs import Unsquashfs, Squashfs
from initrd import InitrdUpdate
from chroot import sudo_chroot
from project import DiskInfo, RemasterProject
from cmdparser import *
from utils import samefile, fileup, InstallationError
from pkglist import PackageList
import utils
import shutil
from uirk.i18n import gettext as _

class Error(Exception):
    pass

class ManageOptionParser(CommandOptionParser):
    COMMANDS = {}
    def __init__(self, *args, **kwargs):
        CommandOptionParser.__init__(self, self.COMMANDS, *args, **kwargs)
        self.description = 'Ubuntu Image Remaster Kit'

    @staticmethod
    def add_command(name, func):
        ManageOptionParser.COMMANDS[name] = func

BASE_OPTIONS = [
    make_option("-f", "--force", action="store_true", dest="force", default=False),
    make_option("-v", "--verbose", action="store_true", dest="verbose", default=False)
]

@register_command(ManageOptionParser, "startproject",
    description=_("Start new UIRK project.\n"
                  "You need to specify ubuntu iso image or path to already mount image.\n"
                  "This operation will take a while.\n"),
    usage="[options] <path-to-iso>",
    args_min=1,
    args_max=1,
    option_list=[
        make_option("--skip-iso", action="store_true", dest="skip_iso", default=False,
                    help=_("Skip extracting ISO image")),
        make_option("--skip-squashfs", action="store_true", dest="skip_squashfs", default=False,
                    help=_("Skip extracting Squashfs image")),
        make_option("--dir", action="store_true", dest="is_dir", default=False,
                    help=_("Specified path is directory and not ISO image")),
        make_option("-p", "--prjdir", dest="prjdir",
                    help=_("Specify project directory, default is current"), metavar="PRJDIR", default="."),
    ] + BASE_OPTIONS)
def start_project(isofile, prjdir=".", is_dir=False, skip_iso=False, skip_squashfs=False, force=False, verbose=False):
    diskdir = os.path.join(prjdir, 'disk')
    if not skip_iso:
        if is_dir:
            if not os.path.isdir(isofile):
                raise Error, _("Specified path is not a directory: %s") % isofile
        else:
            if not os.path.isfile(isofile):
                raise Error, _("Specified path is not a file: %s") % isofile
        if len(os.listdir(prjdir)) > 0 and not force:
            raise Error, _("Project directory is not empty use '--force' to ignore this check")
        ExtractISO(isofile, diskdir, is_dir=is_dir, verbose=verbose).run()
    info = DiskInfo(diskdir)
    prj = RemasterProject(prjdir, diskinfo=info)
    if not skip_squashfs:
        Unsquashfs(os.path.join(prjdir, 'disk/casper/filesystem.squashfs'), os.path.join(prjdir, 'fs'), verbose=verbose).run()
    prj.write()
    prj.update_fs_timestamp()
    print _('Strated new project for %s (%s)') % (prj.label, prj.arch)
    print 'ID: %s' % (prj.distrib_id)
    print 'Codename: %s' % (prj.distrib_codename)
    print 'Releae: %s' % (prj.distrib_release)
    print 'Description: %s' % (prj.distrib_description)
    return prj
start_project.no_project_required=False

SQUASHFS_EXCLUDE = ['tmp/*', 'proc/*', 'sys/*', 'var/cache/apt/archives/*']
SKIP_PACKAGES = ['user-setup', 'lupin-casper', 'casper']

@register_command(ManageOptionParser, 'mksquashfs',
    usage=_("[options]"),
    description=_("Create squashfs, as well as move kernel and initrd to DISK"),
    args_max=0,
    option_list=[
        make_option("-v", "--verbose", action="store_true", dest="verbose", default=False),
        make_option("-f", "--force", action="store_true", dest="force", default=False,
                help=_("Ignore filesystem timestamp, force squashfs image creation"))
    ])
def command_create_squashfs(prj, verbose=False, force=False):
    if not force and not prj.fs_dirty:
        print _('Squashfs image seems to be up to date, use --force switch to continue')
        return
    squash_path = os.path.join(prj.tmp_dir, 'filesystem.squashfs')
    squash_size_path = os.path.join(prj.tmp_dir, 'filesystem.size')

    squash_dest = os.path.join(prj.casper_dir, 'filesystem.squashfs')
    squash_size_dest = os.path.join(prj.casper_dir, 'filesystem.size')


    Squashfs(prj.fsdir, squash_path, squash_size_path, exclude=SQUASHFS_EXCLUDE, wildcards=True).run()

    utils.unlink_safe(squash_dest)
    utils.unlink_safe(squash_size_dest)

    os.rename(squash_size_path, squash_size_dest)
    os.rename(squash_path, squash_dest)

    # trigger timestamp creation
    if not os.path.exists(prj.fstimestamp_path):
        prj.update_fs_timestamp()

    utils.touch(squash_dest)
    utils.touch(squash_size_dest)

    command_kernel_update(prj, verbose=verbose)

    pkgs = list_installed(prj)
    pkgs.sort(key=lambda p: p.name)

    print 'Writting filesystem.manifest'
    fp = open(os.path.join(prj.casper_dir, 'filesystem.manifest'), 'w')
    for pkg in pkgs:
        fp.write('%s %s\n' % (pkg.name, pkg.installedVersion))
    fp.close()
    print 'Writting filesystem.manifest-desktop'
    fp = open(os.path.join(prj.casper_dir, 'filesystem.manifest-desktop'), 'w')
    for pkg in pkgs:
        if pkg.name.startswith('language-support-') or  \
            pkg.name.startswith('language-pack-') or    \
            pkg.name.startswith('ubiquity') or pkg.name in SKIP_PACKAGES:
                continue
        fp.write('%s %s\n' % (pkg.name, pkg.installedVersion))
    fp.close()



def update_md5sums(prj):
    print 'Calculating md5sums...'
    cache = prj.md5cache
    cache.md5sum(prj.diskdir, exclude=['md5sum.txt'])
    cache.write()
    print 'Writing md5sum.txt...'
    fname = os.path.join(prj.diskdir, 'md5sum.txt')
    fp = open(fname, 'w')
    cache.write_md5sums(fp)
    fp.close()

@register_command(ManageOptionParser, "iso",
    usage=_("[path-to-iso-file]"),
    description=_("Create ISO filesystem image, default output is 'out.iso'"),
    args_max=1,
    option_list = BASE_OPTIONS)
def command_create_iso(prj, isofname='out.iso', verbose=False, force=False):
#    from datetime import datetime
#    timestamp = datetime.now().strftime('%Y%m%d')
#    info = os.path.join(prj.diskdir, '.disk/info')
#    fp = open(info, 'w')
#    fp.write('%s - Release %s (%s)\n' % (prj.label, prj.arch, timestamp))
#    fp.close()

    update_md5sums(prj)

## TODO: do we really need this? should be more correct
## .md5cache should written each time we can't relay on file dates
##    if not force and not prj.md5cache.dirty and fileup(isofname, prj.md5cache_path):
##        print "MD5SUMS seems to be okay, use '--force' switch to create iso image anyway"
##        return


    builder = BuildISO(isofname, prj.diskdir, label=prj.label, \
                boot_image='isolinux/isolinux.bin', \
                boot_catalog='isolinux/boot.cat', verbose=verbose)
    builder.run()

@register_command(ManageOptionParser, "shell",
    usage=_("[program] [--] [program options]"),
    description=_("Run shell or program in chrooted target environment."))
def command_shell(prj, *args):
    if len(args):
        cmd = ' '.join(args)
    else:
        cmd = 'bash'
    try:
        sudo_chroot(prj.fsdir, cmd)
    finally:
        prj.update_fs_timestamp()

@register_command(ManageOptionParser, "upgrade",
    description=_("Run apt-get update and upgrade in chrooted target environment."),
    usage="",
    args_max=0,
    option_list=[
        make_option("-y", "--yes", action="store_true", dest="assume_yes",
            default=False, help=_("Assume Yes to all queries and do not prompt")),
        make_option("-U", "--upgrade", action="store_false", dest="do_update",
            default=True, help=_("Do not update apt database, perform upgrade only")),
        make_option("-u", "--update", action="store_false", dest="do_upgrade",
            default=True, help=_("Do not upgrade, perform apt database update only")),
])
def command_upgrade(prj, assume_yes=False, do_upgrade=True, do_update=True):
    cmd = []
    if do_update:
        cmd.append('apt-get update')
    if do_upgrade:
        cmd.append('apt-get %s dist-upgrade' % ("-y" if assume_yes else ""))
    if len(cmd) == 0:
        print _('Nothing to do')
        return
    command_shell(prj, '&&'.join(cmd))

@register_command(ManageOptionParser, "disk-kernel-update",
    description=_("Update disk kernel from target filesystem"),
    usage="",
    option_list=[
        make_option("-v", "--verbose", action="store_true", dest="verbose", default=False)
    ])
def command_kernel_update(prj, verbose=False):
    files = []
    for name in 'initrd.img', 'vmlinuz':
        link = os.path.join(prj.fsdir, name)
        if not os.path.islink(link):
            print "Can not find link to '%s' on target fs" % name
            return
        path = os.readlink(link)
        if not path.startswith('boot'):
            print "Warning! '%s' points to '%s' do not known what to do" % (link, path)
            return
        path = os.path.join(prj.fsdir, path)
        if not os.path.exists(path):
            print "File does not exists '%s'" % path
            return
        files.append(path)

    initrd_src = files[0]
    vmlinuz_src = files[1]

    cached_initrd = os.path.join(prj.tmp_dir, 'initrd')
    cached_uuid = os.path.join(prj.tmp_dir, 'uuid')

    if not samefile(prj.vmlinuz_path, vmlinuz_src):
        print 'Installing new vmlinuz'
        shutil.copy(vmlinuz_src, prj.vmlinuz_path)

    if not samefile(prj.initrd_path, cached_initrd) or  \
            not fileup(cached_initrd, initrd_src) or \
            not samefile(prj.casper_uuid_path, cached_uuid):
        print 'Have to update initrd'
        updater = InitrdUpdate(initrd_src, cached_initrd,
                cached_uuid, prj.initrd_type, verbose=verbose)
        updater.run()
        shutil.copy(cached_initrd, prj.initrd_path)
        shutil.copy(cached_uuid, prj.casper_uuid_path)

def list_installed(prj):
    pkgs = []
    for pkg in prj.apt_cache:
        if pkg.is_installed:
            pkgs.append(pkg)
    return pkgs

@register_command(ManageOptionParser, "installed",
    usage=_("[options]"),
    description=_("List installed packages"))
def command_list_installed(prj):
    pkgs = list_installed(prj)
    pkgs.sort(key=lambda p: p.name)
    for i in pkgs:
        print i.name, i.installedVersion

@register_command(ManageOptionParser, "set-gpg-key",
    usage=_("<gpg-key>"),
    description=_("Set GPG-key that will be used to sign pool packages"),
    args_min=1, args_max=1)
def command_set_gpg_key(prj, gpg_key):
    prj.gpg_key = gpg_key
    prj.write()

def get_pool_deps(prj):
    path = os.path.join(prj.prjdir, 'packages.pool')
    if not os.path.exists(path):
        raise Error, "Please create file 'packages.pool' wich lists packages you want to see in pool"
    lst = PackageList()
    lst.read(path)

    from pool import PoolDeps
    deps = PoolDeps(prj)
    deps.add_pkgnames(lst)
    return deps

@register_command(ManageOptionParser, "pool-fetch",
    usage=_("[options]"),
    description=_("Fetch packages listed in 'packages.pool' file"))
def command_pool_fetch(prj):
    import apt.progress
    from pool import PoolFetcher
    deps = get_pool_deps(prj)
    PoolFetcher(deps).run(progress=apt.progress.TextFetchProgress())

@register_command(ManageOptionParser, "pool-create",
    usage=_("[options]"),
    description=_("Create disk pool tree"),)
def command_pool_create(prj):
    from pool import PoolTreeGenerator
    deps = get_pool_deps(prj)
    PoolTreeGenerator(deps).run()

@register_command(ManageOptionParser, "set-languages",
    usage=_("[locales...]"),
    description=_("Set locales that you would install on your CD"),)
def command_locales(prj, *locales):
    prj.locales = locales
    prj.write()

@register_command(ManageOptionParser, "install-languages",
    usage=_("[options...]"),
    description=_("Install language support files"))
def command_languages_install(prj):
    if 0 == len(prj.locales):
        print _("You did not specified locales you would like to see installed.")
        return

    to_remove = set()
    templates = set(['language-support-%s', 'language-pack-%s'])

    for pkg in list_installed(prj):
        for b in ('language-support-writing',
                    'language-support', 'language-pack',
                    'language-pack-gnome', 'language-pack-kde'):
            if not pkg.name.startswith(b + '-'):
                continue
            parts = pkg.name.split('-')
            if parts[-1] == 'base':
                idx = -2
            else:
                idx = -1
            locale = parts[idx]

            # XXX
            if len(locale) > 3:
                continue

            parts[idx] = '%s'
            templates.add('-'.join(parts))

            if not locale in prj.locales:
                to_remove.add(pkg.name)

    to_install = []
    for locale in prj.locales:
        for t in templates:
            to_install.append(t % locale)
    command_shell(prj, 'apt-get', '-y', 'install', *to_install)
    command_shell(prj, 'apt-get', '-y', 'remove', *to_remove)

@register_command(ManageOptionParser, "install",
    usage=_("[-- options] [packages]"),
    description=_("Install packages into fs"))
def command_install(prj, *args):
    command_shell(prj, 'apt-get', 'install', '-y', *args)

def main():
    parser = ManageOptionParser()
    func, values, args = parser.parse_args()
    try:
        if hasattr(func, 'no_project_required'):
            func(*args, **values.__dict__)
        else:
            prj = RemasterProject('.')
            func(prj, *args, **values.__dict__)
    except Error, e:
        print >> sys.stderr, "Error: %s" % str(e)
        sys.exit(1)
    except InstallationError, e:
        print >> sys.stderr, "InstallationError: %s" % str(e)

if __name__ == "__main__":
    main()
