#  Copyright (c) 2010 Vitja Makarov
#
#  Author: Vitja Makarov <vitja.makarov@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from uirk.utils import md5sum

class CacheItem:
    def __init__(self, path, md5=None, size=0, time=0):
        self.path = path
        self.md5 = md5
        self.size = size
        self.time = time

    def update(self, full_path):
        """Update item's md5sum return True on hit"""
        st = os.stat(full_path)
        size, time = st.st_size, int(st.st_mtime)
        if self.md5 is not None and self.size == size and    \
                self.time == time:
            return True
        old_md5 = self.md5
        self.md5 = md5sum(full_path)
        self.size = size
        self.time = time
        return old_md5 == self.md5

    def __str__(self):
        return '%s %d %d %s' % (self.md5, self.size, self.time, self.path)

class MD5Cache(dict):
    def __init__(self, fname):
        self.fname = fname
        self._dirty_because = set()

        if os.path.exists(fname):
            self._read_cache()

    def _read_cache(self):
        fp = open(self.fname)
        for line in fp:
            if not line:
                continue
            line = line.lstrip().splitlines()[0]
            if line.startswith('#'):
                continue
            try:
                md5, size, time, path = line.split(' ', 3)
            except ValueError:
                continue
            try:
                size = int(size)
                time = int(time)
            except (ValueError, TypeError):
                continue
            self[path] = CacheItem(path, md5, size, time)
        fp.close()

    def _cache_file(self, full_path, path, logger=None):
        if logger:
            logger.info("%s %s", path, self[path].md5)
        if self.has_key(path):
            hit = self[path].update(full_path)
        else:
            self[path] = CacheItem(path)
            self[path].update(full_path)
            hit = False
        if not hit:
            self._dirty_because.add(path)
        return hit

    def write(self):
        fp = open(self.fname, 'w')
        fp.write('# MD5 lazy cache v0\n')
        for i in self.values():
            fp.write('%s\n' % str(i))
        fp.close()

    def md5sum(self, basedir, logger=None, exclude=None):
        for root, dirs, files in os.walk(basedir):
            for f in files:
                full_path = os.path.join(root, f)
                if os.path.islink(full_path):
                    continue
                if callable(exclude):
                    if exclude(full_path):
                        continue
                elif exclude:
                    rel = os.path.relpath(full_path, basedir)
                    if rel in exclude:
                        continue
                path = os.path.relpath(full_path, basedir)
                self._cache_file(full_path, path, logger)

    def __delitem__(self, path):
        if not self.has_key(path):
            return
        dict.__delitem__(self, path)
        if path in self._dirty_because:
            self._dirty_because.remove(path)

    def write_md5sums(self, fp):
        items = self.values()
        items.sort(key=lambda x: x.path)
        for i in items:
            fp.write('%s %s\n' % (i.md5, i.path))

    @property
    def dirty(self):
        return len(self._dirty_because) > 0


if __name__ == "__main__":
    cache = MD5Cache('xxx.md5')
    cache.md5sum(os.path.abspath('./disk/'))
    cache.write()
