#  Copyright (c) 2010 Vitja Makarov
#
#  Author: Vitja Makarov <vitja.makarov@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import fcntl
import resource
import select

from uirk.utils import LineReader

## move me to utils
def smart_exec(name, args, master_fd, fds=(1,2)):
    """Execute program pointed by path `name' with args"""
    fds = list(fds)
    fds.sort()

    ## duplicate status output
    for fd in fds:
        if fd == master_fd:
            continue
        try:
            os.close(fd)
        except OSError:
            pass
        os.dup2(master_fd, fd)

    ## make standard fds point to /dev/null
    devnull = os.open('/dev/null', os.O_RDWR)
    for fd in (0, 1, 2):
        if fd in fds:
            continue
        try:
            os.close(fd)
        except OSError:
            pass
        os.dup2(devnull, fd)

    ## now close all other fds
    maxfd = max(resource.getrlimit(resource.RLIMIT_NOFILE))
    startfd = 3
    for fd in fds:
        os.closerange(startfd, fd)
        startfd = fd + 1
    os.closerange(startfd, maxfd)

    ## and now execute
    try:
        os.execv(name, args)
    except OSError, e:
        print e
    os._exit(666)

class BaseProgress(object):
    def __init__(self, obj=None):
        if obj is None:
            raise AttributeError, \
                "'obj' parameter must be set to create BaseProgress instance"
        self.obj = obj

    def handle_timeout(self):
        pass
    def error(self, error):
        pass
    def pulse(self):
        pass

class BaseSpawnProgress(BaseProgress):
    select_timeout = 0.5

    def __init__(self, root_required=False, verbose=False, *args, **kwargs):
        self.root_required = root_required
        self.verbose = verbose
        (self.statusfd, self.writefd) = os.pipe()
        self.write_stream = os.fdopen(self.writefd, "w")
        self.status_stream = os.fdopen(self.statusfd, "r")
        self.line_reader = LineReader(self.status_stream)
        if not hasattr(self, 'sudo_cmd'):
            self.sudo_cmd = None
        fcntl.fcntl(self.statusfd, fcntl.F_SETFL, os.O_NONBLOCK)
        super(BaseSpawnProgress, self).__init__(*args, **kwargs)

    def run(self, obj, prog, args):
        child_pid = self.fork()
        if child_pid == 0:
            self.spawn(prog, args, root_required=self.root_required)
        return self.wait_child(child_pid)

    def fork(self):
        return os.fork()

    def spawn(self, name, args, fds = (1, 2), root_required=False):
        if root_required:
            if self.sudo_cmd is None:
                raise AttributeError, "root_required is set but no sudo_cmd specified"
            args = [self.sudo_cmd] + args
            name = self.sudo_cmd
        smart_exec(name, args, self.writefd, fds)

    def wait_child(self, child_pid):
        """Wait for child progress to exit.

        This method is responsible for calling update_interface() from time to
        time. It exits once the child has exited. The return values is the
        full status returned from os.waitpid() (not only the return code).
        """
        (pid, res) = (0, 0)
        while True:
            try:
                select.select([self.status_stream], [], [],
                              self.select_timeout)
            except select.error, (errno_, _errstr):
                if errno_ != errno.EINTR:
                    raise
            if not self.update_interface():
                self.handle_timeout()
            else:
                self.pulse()

            try:
                (pid, res) = os.waitpid(child_pid, os.WNOHANG)
                if pid == child_pid:
                    break
            except OSError, err:
                if err.errno == errno.ECHILD:
                    break
                if err.errno != errno.EINTR:
                    raise
        if res:
            self.error(res)
        else:
            self.complete()
        return res

    def update_interface(self):
        ret = False
        while True:
            try:
                line = self.line_reader.readline()
            except IOError, err:
                return ret
            if line is None:
                return
            self.handle_status_line(line)
            ret = True

    def log_verbose(self, data):
        pass

    def handle_status_line(self, line):
        """This function is called when new line arrived to status pipie"""
        self.log_verbose(line)

    def complete(self):
        """Called on success"""

    def failed(self, res):
        """Called when subprocess exited with non-zero code or killed by signal"""

class BaseBuildISOProgress(BaseSpawnProgress):
    """BuildISO progress handler"""
    def __init__(self, *args, **kwargs):
        super(BaseBuildISOProgress, self).__init__(*args, **kwargs)

    def handle_status_line(self, line):
        self.log_verbose(line)
        line = line.strip().split()

        if len(line) == 0:
            return
        if line[0].endswith('%'):
            try:
                percents = float(line[0][:-1])
                self.update_percents(percents)
            except ValueError:
                pass

    def update_percents(self, percents):
        """This function is called when updated percents information is updated"""

class BaseSquashfsProgress(BaseSpawnProgress):
    def __init__(self, unsquashfs=False, *args, **kwargs):
        kwargs['root_required'] = True
        super(BaseSquashfsProgress, self).__init__(*args, **kwargs)
        self._unsquashfs = unsquashfs
        self.line_reader.add_separator('\r')

    def handle_status_line(self, line):
        self.log_verbose(line)

        line = line.strip().split()

        if len(line) == 0:
            return

        if line[-1].endswith('%'):
            try:
                percents = float(line[-1][:-1])
                self.update_percents(percents)
            except ValueError:
                pass

    def update_percents(self, percents):
        """This function is called when updated percents information is updated"""

class BaseExtractISOProgress(BaseSpawnProgress):
    """ExtractISO progress handler"""
    def __init__(self, *args, **kwargs):
        kwargs['root_required'] = True
        super(BaseExtractISOProgress, self).__init__(*args, **kwargs)

    def handle_status_line(self, line):
        """We don't have percent information here"""

class BaseInitrdUpdateProgress(BaseSpawnProgress):
    def __init__(self, *args, **kwargs):
        super(BaseInitrdUpdateProgress, self).__init__(*args, **kwargs)

class BaseExecTaskProgress(BaseSpawnProgress):
    def __init__(self, *args, **kwargs):
        self.message = kwargs.pop('message', 'unknown')
        super(BaseExecTaskProgress, self).__init__(*args, **kwargs)
