#  Copyright (c) 2010 Vitja Makarov
#
#  Author: Vitja Makarov <vitja.makarov@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import os.path
import hashlib

_cached_path = os.getenv('PATH', '/usr/bin:/usr/local/bin').split(':')

PROJECT_ROOT = os.path.normpath(os.path.dirname(__file__))

if PROJECT_ROOT[0:1] != '/':
    PROJECT_ROOT = os.path.join(os.getcwd(), PROJECT_ROOT)

SCRIPTS_DIR = os.path.join(PROJECT_ROOT, 'scripts')
OWNERFIX_SCRIPT = os.path.join(SCRIPTS_DIR, 'ownerfix.sh')

def which(name, path=None):
    if path is None:
        path = _cached_path
    for i in path:
        prog = os.path.join(i, name)
        if os.access(prog, os.X_OK):
            return prog

SUDO = which('sudo')
GKSUDO = which('gksudo')

class LineReader(object):
    """Read lines from fp"""
    def __init__(self, fp, separators=None):
        self.fp = fp
        if separators is None:
            self.separators = ['\n']
        else:
            self.separators = separators
        self.buf = ''

    def add_separator(self, sep):
        self.separators.append(sep)

    def readline(self):
        while True:
            idx = []
            for sep in self.separators:
                i = self.buf.find(sep)
                if i < 0:
                    continue
                idx.append(i + len(sep))
            if len(idx):
                idx = min(idx)
                line = self.buf[:idx]
                self.buf = self.buf[idx:]
                return line
            data = self.fp.read(4096)
            if not len(data):
                return
            self.buf += data

    def __iter__(self):
        while True:
            line = self.readline()
            if line is None:
                break
            yield line

class InstallationError(Exception):
    """Raised when some external tool is not found"""

class OperationError(Exception):
    """Operation failed"""

class RootRequired(Exception):
    """Raised when root privileges are required"""

def is_root():
    return os.getuid() == 0

def require_root(func):
    def decorator(*args, **kwargs):
        if is_root():
            raise RootRequired, "Root privileges are required, did you miss sudo?"
        return func(*args, **kwargs)
    return decorator

def never_giveup(func):
    def decor(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        finally:
            os._exit(666)
    return decor

class cached_property(object):
    def __init__(self, func, cache_to=None, doc=None):
        self.cache_to = cache_to or '__cached_%s' % func.__name__
        self.func = func
        self.__doc__ = doc

    def __get__(self, instance, objtype=None):
        if instance is None:
            return self
        if not hasattr(instance, self.cache_to):
            setattr(instance, self.cache_to, self.func(instance))
        return getattr(instance, self.cache_to)

    def __set__(self, instance, value):
        raise AttributeError, "can't set cachable attribute"

    def __delete__(self, instance):
        """Delete cached value, force value update on next access"""
        try:
            delattr(instance, self.cache_to)
        except AttributeError:
            pass

try:
    from apt_pkg import md5sum as _md5sum
    def md5sum(fname):
        fp = open(fname)
        md5 = _md5sum(fp)
        fp.close()
        return md5
except ImportError:
    def md5sum(fname):
        m = hashlib.md5()
        fp = open(fname, 'rb')
        while True:
            data = fp.read(4096)
            if not data:
                break
            m.update(data)
        fp.close()
        return m.hexdigest()

def samefile(a, b):
    """Returns True if both files are the same"""
    try:
        sta = os.stat(a)
        stb = os.stat(b)
        if sta.st_size != stb.st_size:
            return
    except OSError:
        return False
    return md5sum(a) == md5sum(b)

def fileup(a, b):
    """Returns true if file a date is younger than file b"""
    try:
        sta = os.stat(a)
        stb = os.stat(b)
    except OSError:
        return False
    return sta.st_mtime > stb.st_mtime

def mkdir_safe(path):
    """Make directory and do not raise exception if it already exists"""
    try:
        os.mkdir(path)
    except OSError, e:
        if e.errno != os.errno.EEXIST:
            raise

def makedirs_safe(path, mode=0777):
    """Like mkdir -p"""
    if os.path.isdir(path):
        return
    os.makedirs(path, mode)

def unlink_safe(path):
    try:
        os.unlink(path)
    except OSError, e:
        if e.errno != os.errno.ENOENT:
            raise

def same_inode(a, b):
    sta, stb = os.stat(a), os.stat(b)
    return sta.st_ino == stb.st_ino

def touch(fname):
    fp = open(fname, 'a')
    fp.close()
